/*********************************************************************************
 * libcapifax.c
 *
 * Copyright (c) 2008-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "libcapifax.h"
#include "alaw_ulaw.h"

static capifax_con_t        connection[CAPI_CONNECTIONS];
static _cmsg                cMsg;
static int                  ApplId = -1;
static int                  in_loop = 0;
static int                  in_thread = 0;
static int                  boosted = 0;
static int                  log_level = 0;
static int                  bchannels = 0;
static int                  controllers = 0;
static int8_t               *_linear16_2_law = &linear16_2_law[32768];
static uint16_t             *_law_2_linear16 = &law_2_linear16[0];
static capifax_status       ev_status = NULL;
static capifax_progress     ev_progress = NULL;
static capifax_log          ev_log = NULL;
static capifax_receive      ev_receive = NULL;

static int capi_init(int controller);
static int capi_close();
static int capi_loop();
static int capi_ind();
static void capi_conf();
static void capi_rx_tx(int con);
static void capi_get_source_no(_cmsg *CMSG, char number[256]);
static void capi_get_target_no(_cmsg *CMSG, char number[256]);
static void capi_resp_connection(int PLCI, unsigned int ignore);
static void capi_term_connection(int con);
static int capi_call(const char *src_no, const char *trg_no, unsigned controller, unsigned call_anonymous);
static int capi_get_free();
static int capi_set_free(int con);
static int capi_find_plci(int PLCI);
static int capi_find_ncci(int NCCI);
static void capi_error(long info);

static int spandsp_init(const char *tiff_file, int calling_party, int modem, int ecm, const char *LSI, const char *LocalHeaderInfo, int conid);
static int spandsp_close(int con);
static int spandsp_rx(int con, uint8_t *buf, size_t len);
static int spandsp_tx(int con, uint8_t *buf, size_t len);

#if SPANDSP_VERSION >= 5
static int      phase_b_handler (t30_state_t *s, void *user_data, int result);
static int      phase_d_handler (t30_state_t *s, void *user_data, int result);
#else
static void     phase_b_handler (t30_state_t *s, void *user_data, int result);
static void     phase_d_handler (t30_state_t *s, void *user_data, int result);
#endif
#if SPANDSP_VERSION >= 6
static void real_time_frame_handler(t30_state_t *s, void *user_data, int direction, const uint8_t *msg, int len);
#endif

static void     phase_e_handler (t30_state_t *s, void *user_data, int result);
static void     spandsp_msg_log(int level, const char *text);

static void spandsp_msg_log(int level, const char *text);
static void dbg(int level, const char *format, ...);
static void boost_priority(int boost);

/* --------------------------------------------------------
      EXPORTS 
  --------------------------------------------------------*/

// Send Fax
//
// tiff_file: The Tiff file to send
// modem: 0-3 (2400-14400)
// ecm: Error correction mode (on/off)
// controller: The controller for sending the fax
// src_no: MSN
// dst_no: Target fax number
// LSI: Fax ident
// LocalHeaderInfo: Fax header line
// call_anonymous: Send fax anonymous 
//
EXTERN int capifax_send(const char *tiff_file, int modem, int ecm, int controller, const char *src_no, const char *trg_no, const char *LSI, const char *LocalHeaderInfo, int call_anonymous)                       
{    
    int con = capi_get_free();
    int ret = -2;
    
    dbg(1, "capifax_send(tiff: %s, modem: %d, ecm: %s, controller: %d, src: %s, trg: %s, ident: %s, header: %s, anonymous: %d)", tiff_file, modem, ecm ? "on" : "off", controller, src_no, trg_no, (LSI != NULL ? LSI : "(null)"), (LocalHeaderInfo != NULL ? LocalHeaderInfo : "(null)"), call_anonymous);
            
    if(con != -1)
    {   
        connection[con].status.phase = IDLE;
        connection[con].PLCI = -2;
        connection[con].status.error_code = 0;     
        connection[con].status.capi_code = 0;
        connection[con].status.sending = 1;
        connection[con].status.manual_hookup = 0;   
        connection[con].status.modem = modem;
        connection[con].status.ecm = ecm;
        memset(connection[con].status.ident, 0, sizeof(connection[con].status.ident));
        memset(connection[con].status.header, 0, sizeof(connection[con].status.header));
        memset(connection[con].status.src_no, 0, sizeof(connection[con].status.src_no));
        memset(connection[con].status.trg_no, 0, sizeof(connection[con].status.trg_no));
        memset(connection[con].status.tiff_file, 0, sizeof(connection[con].status.tiff_file));

        snprintf(connection[con].status.header, sizeof(connection[con].status.header), "%s", LocalHeaderInfo);
        snprintf(connection[con].status.ident, sizeof(connection[con].status.ident), "%s", LSI);
        snprintf(connection[con].status.src_no, sizeof(connection[con].status.src_no), "%s", src_no);
        snprintf(connection[con].status.trg_no, sizeof(connection[con].status.trg_no), "%s", trg_no);
        snprintf(connection[con].status.tiff_file, sizeof(connection[con].status.tiff_file), "%s", tiff_file);                

        spandsp_init(connection[con].status.tiff_file, TRUE, connection[con].status.modem, connection[con].status.ecm, connection[con].status.ident, connection[con].status.header, con);    
        ret = capi_call(src_no, trg_no, (unsigned)controller, (unsigned int)call_anonymous);
        
        if(!ret) 
        {
            if(!capi_loop()) {
                ret = connection[con].status.error_code;                                                
                spandsp_close(con);                
            }
        } 
        else 
        {
            capi_set_free(con);
            spandsp_close(con);
            con = -1;            
        }                
    }
    
    // return con_id when in thread (capi_loop)
    if(in_thread) ret = con;

    return ret;
}

// Receive Fax
//
// tiff_file: The Tiff file for saving the fax
// modem: 0-3 (2400-14400)
// ecm: Error correction mode (on/off)
// src_no: MSN
// dst_no: After receiving a fax, dst_no is the senders fax number
// manual_hookup: Hook up manually
//
// LSI and LocalHeaderInfo are unused!
//
EXTERN int capifax_recv(const char *tiff_file, int modem, int ecm, const char *src_no, char *trg_no, const char *LSI, const char *LocalHeaderInfo, int manual_hookup)
{
    int con = capi_get_free();
    int ret = -2;

    dbg(1, "capifax_recv(tiff: %s, modem: %d, ecm: %s, src: %s, manual: %s)", tiff_file, modem, ecm ? "on" : "off", src_no, manual_hookup ? "on" : "off");
            
    if(con != -1)
    {   
        connection[con].status.phase = IDLE;
        connection[con].status.sending = 0;
        connection[con].status.modem = modem;
        connection[con].status.ecm = ecm;        
        connection[con].status.manual_hookup = manual_hookup;
        connection[con].status.error_code = 0; 
        connection[con].status.capi_code = 0;
        connection[con].PLCI = -1;
        
        memset(connection[con].status.ident, 0, sizeof(connection[con].status.ident));
        memset(connection[con].status.header, 0, sizeof(connection[con].status.header));
        memset(connection[con].status.src_no, 0, sizeof(connection[con].status.src_no));
        memset(connection[con].status.trg_no, 0, sizeof(connection[con].status.trg_no));
        memset(connection[con].status.tiff_file, 0, sizeof(connection[con].status.tiff_file));
        
        snprintf(connection[con].status.src_no, sizeof(connection[con].status.src_no), "%s", src_no);
        snprintf(connection[con].status.tiff_file, sizeof(connection[con].status.tiff_file), "%s", tiff_file);
                
        spandsp_init(connection[con].status.tiff_file, FALSE, connection[con].status.modem, connection[con].status.ecm, connection[con].status.ident, connection[con].status.header, con);    

        if(!capi_loop()) {
            ret = connection[con].status.error_code;                            
            spandsp_close(con);                
        } else {
            ret = -1;
        }

        snprintf(trg_no, sizeof(connection[con].status.trg_no), "%s", connection[con].status.trg_no);
    }
    
    // return con_id when in thread (capi_loop)
    if(in_thread) ret = con;

    return ret;
}

// Fax-Thread Loop 
// When this function runs in application thread, capifax_send() and capifax_recv() returns immediately.
// So it is possible to register multiple accounts for receiving and sending 
//
EXTERN int capifax_loop()
{
    int ret = -1;

    in_thread = 1;
    ret = capi_loop();    
    in_thread = 0;

    return ret;
}


// Initialize libcapifax
//
// This function can be called multiple times to register more controllers
//
// controller: The controller, listen to (receiving)
// loglevel: Loglevel output (0-4)
// status_event: (Optional) callback to status event function
// progress_event: (Optional) callback to progress event function
// log_event: (Optional) callback to log event function
//
EXTERN int capifax_init(int controller, int loglevel, capifax_status status_event, capifax_progress progress_event, capifax_log log_event, capifax_receive receive_event)
{
    int ret = 0;
    int first_init = (ApplId == -1) ? 1 : 0;
    
    log_level = loglevel;
    ev_status = status_event;
    ev_progress = progress_event;
    ev_log = log_event;
    ev_receive = receive_event;

    if(first_init)
        dbg(1, "%s %s %s", LIBNAME, LIBVER, LIBCR);        

    ret = capi_init(controller);            
    if(!ret) create_table_buffer(CAPI_ULAW);
    in_loop = 0;
    in_thread = 0;

    dbg(2, "capifax_init(ctrl: %d, loglevel: %d, ev_status: %x, ev_progress: %x, ev_log: %x)", controller, log_level, &ev_status, &ev_progress, &ev_log);

    if(!ret) {        
        if(first_init) {            
            dbg(1, "%s successfully initialized!", LIBNAME);        
        }
    } else {
        dbg(1, "%s initialization failed! Error %d!", LIBNAME, ret);
    }
    
    return ret;
}

// Destroy libcapifax 
//
EXTERN int capifax_close()
{    
    capi_close();    
    in_loop = 0;
    in_thread = 0;
    spandsp_close(-1);

    dbg(1, "%s %s closed!", LIBNAME, LIBVER);
    
    return 0;
}

// Manually disconnect a specific connection
//
EXTERN int capifax_terminate(int con)
{    
    int ret = -1;
    
    connection[con].status.error_code = 0x1000;  // user aborted
    capi_term_connection(con);
    
    return ret;
}

// Manually accept a specific connection (hook up)
//
EXTERN int capifax_accept(int con)
{
    int ret = -1;

    if(connection[con].PLCI > 0 && connection[con].NCCI <= 0)
    {
        capi_resp_connection(connection[con].PLCI, 0);
        ret = 0;
    }

    return ret;
}

// Manually reject a specific connection 
//
EXTERN int capifax_reject(int con) 
{
    int ret = -1;

    if(connection[con].PLCI > 0 && connection[con].NCCI <= 0)
    {
        connection[con].status.phase = FAX_REJECTED;
        connection[con].status.error_code = 0x999;  // user rejected
        capi_resp_connection(connection[con].PLCI, 1);        
        ret = 0;
    }

    return ret;

}


/* --------------------------------------------------------
    CAPI functions
-------------------------------------------------------- */
static int capi_init(int controller)
{
    CAPI_REGISTER_ERROR ErrorCode = 0;    
    unsigned int _ApplId = -1;    
    int i;
    int start = 0;
    int end = 0;
    int skip = -1;
    
    // Test CAPI
    if(bchannels == 0 && controllers == 0 && ApplId == -1)
    {
        struct capi_profile capiprofile;
        
        #ifdef WIN32        
            _ApplId = CAPI20_REGISTER(CAPI_BUFFERCNT * (CAPI_PACKETS * 2), 1, CAPI_BUFFERCNT, CAPI_PACKETS, &ErrorCode);
        #else
            ErrorCode = CAPI20_REGISTER(1, CAPI_BUFFERCNT, CAPI_PACKETS, &_ApplId);
        #endif
	    if(ErrorCode) {        
        	_ApplId = (unsigned short)-1;        
        	dbg(1, "Error: CAPI not installed!");
        	return -1;
	    }

        // Fetch controller/bchannel count        
        CAPI20_GET_PROFILE(0, (unsigned char *)&capiprofile);        
        controllers = capiprofile.ncontroller;

        for(i=1; i<=capiprofile.ncontroller; i++)
        {
            struct capi_profile capiprofile2; 
		 	CAPI20_GET_PROFILE(i, (unsigned char *)&capiprofile2);
            char *pnManu = capiprofile2.manu;
            bchannels += capiprofile2.nbchannel;            
            dbg(2, "Found controller #%d with %d B-channel(s)", i, capiprofile2.nbchannel);
            if ( capiprofile2.nbchannel == 2 && pnManu[ 3 ] == 0x00 ) {
	            dbg(2, "-> skipping internal controller %d", i);
	            skip = i;
	        }
        }

        dbg(1, "Found %d controllers!", capiprofile.ncontroller);

        // Release CAPI
        CAPI20_RELEASE(_ApplId);
        _ApplId = -1;

        // Reset connection struct
        memset(&connection, 0, sizeof(connection));
    }
    
    // Listen to all (<=0) or single controller (>=1)
    if(controller <= 0)
    {
        start = 1; 
        end = controllers;        
    }     
    else
    {
        start = controller; 
        end = controller;
    }
    
    // Register with CAPI
    if(ApplId == -1) 
    {        
        #ifdef WIN32
        _ApplId = (unsigned int)CAPI20_REGISTER(CAPI_BUFFERCNT * (CAPI_PACKETS * 2), CAPI_BCHANNELS, CAPI_BUFFERCNT, CAPI_PACKETS, &ErrorCode);
        #else
        ErrorCode = CAPI20_REGISTER(CAPI_BCHANNELS, CAPI_BUFFERCNT, CAPI_PACKETS, &_ApplId);
        #endif
        if(ErrorCode) {
            dbg(1, "Error: CAPI registration error!");
            return -2;    // registration error!
        }

        ApplId = _ApplId;
    }

    // Listen to CAPI controller(s)
    for(i=start; i<=end; i++) {
    	if ( controller <= 0 && skip == i ) {
    		// if we are using auto channels, skip internal controller
    		continue;
    	}
        LISTEN_REQ(&cMsg, ApplId, 0, i, 0x3ff, 0x1FFF03FF, 0, NULL, NULL);
        dbg(1, "Listen to controller #%d ...", i);
    }
    
    return 0; // ok!
}

static int capi_close()
{
    int i;

    if(ApplId != -1) {
	for(i=0; i<CAPI_CONNECTIONS; i++) {
		capifax_terminate(i);
		usleep(25);
	}

        in_loop = 0;
        in_thread = 0;
	bchannels = 0;

	usleep(250);
	CAPI20_RELEASE(ApplId);        
    }
	
    ApplId = -1;
    return 0;
}

static int capi_loop()
{    
    struct timeval tv;    
    unsigned int Info;    
    
    // already in loop? return!
    if(in_loop) { in_loop++; return in_loop; }
    in_loop++;
                
    while(in_loop)
    {
	tv.tv_sec = 1;
	tv.tv_usec = 0;

        if(CAPI20_WaitforMessage(ApplId, &tv) == CapiNoError) {
            switch(Info = capi_get_cmsg(&cMsg, ApplId))
            {   
                case 0x0000:
                    switch(cMsg.Subcommand)
                    {
                        // Indication
                        case CAPI_IND:
                            capi_ind();
                            break;

                        // Confirmation
                        case CAPI_CONF:
                            capi_conf();
                            break;
                    }            
                    break;

                default:
                    usleep(1);  // prevent CPU hog
                    break;
            }
        }
	else
	  usleep(1);
    }

    return 0;
}

static int capi_ind()
{
    _cmsg CMSG1;    
    int PLCI = -1;
    int NCCI = -1;        
    int iMsgNr = 0;
    int tmp_conid = -1;
    char SourcePhoneNumber[256];
    char TargetPhoneNumber[256];            
    int i;
    int con = -1;
    int cip = -1;

    // Get Msg-Nr. & Controller
    iMsgNr = cMsg.Messagenumber;

    switch(cMsg.Command)
    {
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_CONNECT - Connect
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_CONNECT:
        PLCI = CONNECT_IND_PLCI(&cMsg);
        capi_get_source_no(&cMsg, SourcePhoneNumber);
        capi_get_target_no(&cMsg, TargetPhoneNumber);
        cip = CONNECT_IND_CIPVALUE( &cMsg );

        // Already in queue?
        tmp_conid = capi_find_plci(PLCI);

        // Manually receive event (can be called by external application)
        if(tmp_conid == -1 && cip != 0 && ev_receive != NULL) 
        {
            char *file;
            tmp_conid = capi_get_free();

            if(tmp_conid != -1) 
            {
                if(!ev_receive(con, TargetPhoneNumber, SourcePhoneNumber, &file)) 
                {
			        spandsp_init(file, FALSE, 3, TRUE, "+00 0000 00000", "libcapifax", tmp_conid);                

		            connection[tmp_conid].status.phase = IDLE;
		            connection[tmp_conid].status.sending = 0;
		            connection[tmp_conid].status.modem = 3;
		            connection[tmp_conid].status.ecm = FALSE;
		            connection[tmp_conid].status.manual_hookup = 0;
		            connection[tmp_conid].status.error_code = 0;        
		            connection[tmp_conid].PLCI = -1;
        
		            memset(connection[tmp_conid].status.ident, 0, sizeof(connection[tmp_conid].status.ident));
		            memset(connection[tmp_conid].status.header, 0, sizeof(connection[tmp_conid].status.header));
		            memset(connection[tmp_conid].status.src_no, 0, sizeof(connection[tmp_conid].status.src_no));
		            memset(connection[tmp_conid].status.trg_no, 0, sizeof(connection[tmp_conid].status.trg_no));
		            memset(connection[tmp_conid].status.tiff_file, 0, sizeof(connection[tmp_conid].status.tiff_file));

		            snprintf(connection[tmp_conid].status.src_no, sizeof(connection[tmp_conid].status.src_no), "%s", TargetPhoneNumber);
		            snprintf(connection[tmp_conid].status.tiff_file, sizeof(connection[tmp_conid].status.tiff_file), "%s", file);                   	
                }
            }
        }

        // Search for free connection (and check if MSN is correct)
        if(tmp_conid == -1) {
            for(i=0; i<CAPI_CONNECTIONS; i++)
            {
                if(connection[i].PLCI == -1 && 
                   !connection[i].status.sending &&
                   !strcmp(TargetPhoneNumber, connection[i].status.src_no)) {
                    tmp_conid = i;
                    break;
                }
            }
        }
        
        if(tmp_conid == -1)
        {                                                
            // Ignore
            dbg(1, "IND: CAPI_CONNECT - CON: %d, PLCI: %d, NCCI: %d - IGNORING (%s <- %s)", con, PLCI, 0, TargetPhoneNumber, SourcePhoneNumber);        
            capi_resp_connection(PLCI, 1);
            break;
        }

        con = tmp_conid;

        snprintf(connection[con].status.trg_no, sizeof(connection[con].status.trg_no), "%s", SourcePhoneNumber);

        if(!connection[con].status.sending && !strcmp(TargetPhoneNumber, connection[con].status.src_no)) 
        {               
            // Respond
            if(!connection[con].status.manual_hookup) 
            {
                // automatic hook up                
                dbg(1, "IND: CAPI_CONNECT - CON: %d, PLCI: %d, NCCI: %d - RESPONDING (%s <- %s)", con, PLCI, 0, TargetPhoneNumber, SourcePhoneNumber);        
                capi_resp_connection(PLCI, 0);                
            } 
            else 
            {
                // manual hook up
                // Alert
                ALERT_REQ(&CMSG1, ApplId, 0, PLCI,
                        NULL,       
                        NULL,       
                        NULL,       
                        NULL, 
                        NULL);                
            }

            // Event: "CONNECT"
            if(connection[con].status.phase == IDLE && connection[con].PLCI == -1) 
            {
                connection[con].status.phase = CONNECT;
                connection[con].status.bytes_received = 0;
                connection[con].status.bytes_sent = 0;
                connection[con].status.bytes_total = 0;
                connection[con].PLCI = PLCI;
                                
                if(ev_status) ev_status(con, &connection[con].status);            
            }                         
        } 
        else 
        {                                                
            // Ignore
            dbg(1, "IND: CAPI_CONNECT - CON: %d, PLCI: %d, NCCI: %d - IGNORING (%s <- %s)", con, PLCI, 0, TargetPhoneNumber, SourcePhoneNumber);
            capi_resp_connection(PLCI, 1);
        }
        break;

    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_CONNECT_ACTIVE - Active
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_CONNECT_ACTIVE:
        PLCI = CONNECT_IND_PLCI(&cMsg);                
        CONNECT_ACTIVE_RESP(&CMSG1, ApplId, (_cword)iMsgNr, PLCI);
                
        tmp_conid = capi_find_plci(PLCI);        
        if(tmp_conid == -1) break;
        con = tmp_conid;
        
        // Request B3 when sending...
        if(connection[con].status.sending)
            CONNECT_B3_REQ(&CMSG1, ApplId, 0, PLCI, 0);
                
        dbg(2, "IND: CAPI_CONNECT_ACTIVE - CON: %d, PLCI: %d, NCCI: %d", con, connection[con].PLCI, connection[con].NCCI);        
        break;
    
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_CONNECT_B3 - Data connect
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_CONNECT_B3:
        NCCI = CONNECT_B3_IND_NCCI(&cMsg);
        PLCI = NCCI & 0x0000ffff;      
        
        tmp_conid = capi_find_plci(PLCI);
        if(tmp_conid == -1) break;
        else con = tmp_conid;
        
        connection[con].NCCI = NCCI;
        CONNECT_B3_RESP(&CMSG1, ApplId, (_cword)iMsgNr, NCCI, 0, (_cstruct)NULL);
        
        dbg(2, "IND: CAPI_CONNECT_B3 - CON: %d, PLCI: %d, NCCI: %d", con, connection[con].PLCI, connection[con].NCCI);        
        break;

    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_CONNECT_B3_ACTIVE - Data active
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_CONNECT_B3_ACTIVE:
        NCCI = CONNECT_B3_IND_NCCI(&cMsg);
        PLCI = NCCI & 0x0000ffff;        
        
        tmp_conid = capi_find_plci(PLCI);
        if(tmp_conid == -1) break;
        else con = tmp_conid;        
        
        if(PLCI == connection[con].PLCI) connection[con].NCCI = NCCI;
        CONNECT_B3_ACTIVE_RESP(&CMSG1, ApplId, (_cword)iMsgNr, NCCI);
        
        // boost priority
        boost_priority(1);

        dbg(2, "IND: CAPI_CONNECT_B3_ACTIVE - CON: %d, PLCI: %d, NCCI: %d", con, connection[con].PLCI, connection[con].NCCI);        
        break;

    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_DATA_B3 - Data - receive/send
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_DATA_B3:
        NCCI = CONNECT_B3_IND_NCCI(&cMsg);
        PLCI = NCCI & 0x0000ffff;
        
        tmp_conid = capi_find_ncci(NCCI);
        if(tmp_conid == -1) break;
        else con = tmp_conid;
        
        dbg(4, "IND: CAPI_DATA_B3 - CON: %d, PLCI: %d, NCCI: %d", con, connection[con].PLCI, connection[con].NCCI);
        capi_rx_tx(con);
        break;

    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_FACILITY - Facility (DTMF)
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_FACILITY:
        NCCI = CONNECT_B3_IND_NCCI(&cMsg);
        PLCI = NCCI & 0x0000ffff;
        FACILITY_RESP(&CMSG1, ApplId, iMsgNr, PLCI, FACILITY_IND_FACILITYSELECTOR(&cMsg), FACILITY_IND_FACILITYINDICATIONPARAMETER(&cMsg));
        
        tmp_conid = capi_find_ncci(NCCI);
        if(tmp_conid == -1) break;
        else con = tmp_conid;
        
        dbg(2, "IND: CAPI_FACILITY - CON: %d, PLCI: %d, NCCI: %d", con, connection[con].PLCI, connection[con].NCCI);
        break;

    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_INFO
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_INFO:
        PLCI = INFO_IND_PLCI(&cMsg);
        
        // Respond to INFO
        INFO_RESP(&CMSG1, ApplId, iMsgNr, PLCI);

        switch(INFO_IND_INFONUMBER(&cMsg))
        {
             case 0x8045:    // Disconnect  
                 tmp_conid = capi_find_plci(PLCI);
                 if(tmp_conid == -1) break;
                 else con = tmp_conid;

                 capi_term_connection(con);                 
                 break;
        }
        break;


    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_DISCONNECT_B3 - Disconnect data
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_DISCONNECT_B3:
        NCCI = CONNECT_B3_IND_NCCI(&cMsg);
        PLCI = NCCI & 0x0000ffff;        
        DISCONNECT_B3_RESP(&CMSG1, ApplId, iMsgNr, NCCI);
        
        tmp_conid = capi_find_ncci(NCCI);
        if(tmp_conid == -1) capi_find_plci(PLCI);
        if(tmp_conid == -1) break;
        else con = tmp_conid;

        connection[con].NCCI = 0;
        capi_term_connection(con);

        dbg(2, "IND: CAPI_DISCONNECT_B3 - CON: %d, PLCI: %d, NCCI: %d", con, connection[con].PLCI, connection[con].NCCI);    
        break;

    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    // CAPI_DISCONNECT - Disconnect
    // ------------------------------------------------------------------------------------------------------------------------------------------------------
    case CAPI_DISCONNECT:
        PLCI = CONNECT_IND_PLCI(&cMsg);         
        DISCONNECT_RESP(&CMSG1, ApplId, iMsgNr, PLCI);
        capi_error(DISCONNECT_IND_REASON(&cMsg));
                        
        tmp_conid = capi_find_plci(PLCI);
        if(tmp_conid == -1) break;
        else con = tmp_conid;
            
        // CAPI-Error code
        connection[con].status.capi_code = DISCONNECT_IND_REASON(&cMsg);
        
        dbg(2, "IND: CAPI_DISCONNECT - CON: %d, PLCI: %d, NCCI: %d", con, connection[con].PLCI, connection[con].NCCI);    
        capi_set_free(con);
        return 1;
    }

    return 0;
}

static void capi_conf()
{    
    int tmp_plci, tmp_conid;
    
    //con.iMsgNr = con.cMsg->Messagenumber;
    //con.controller = con.cMsg->adr.adrController & 0x7f;  // convert from __u32-to-int

    switch(cMsg.Command)
    {
        case CAPI_FACILITY:        
            dbg(2, "CNF: CAPI_FACILITY");
            capi_error(cMsg.Info);
            break;

        case CAPI_LISTEN:
            dbg(2, "CNF: CAPI_LISTEN");
            capi_error(cMsg.Info);
            break;

        case CAPI_ALERT:
            dbg(2, "CNF: CAPI_ALERT");
            capi_error(cMsg.Info);
            break;

        case CAPI_DATA_B3:
            dbg(4, "CNF: CAPI_DATA_B3");
            capi_error(cMsg.Info);
            break;

        case CAPI_INFO:
            dbg(2, "CNF: CAPI_INFO");
            capi_error(cMsg.Info);
            break;

        case CAPI_CONNECT:
            tmp_plci = CONNECT_IND_PLCI(&cMsg);
                    
            dbg(2, "CNF: CAPI_CONNECT - (PLCI: %d)", tmp_plci);        
            tmp_conid = capi_find_plci(-2);                    // .. or new outgoing call? get plci.
            if(tmp_conid != -1) 
            { 
                // Status: CONNECT            
                connection[tmp_conid].PLCI = CONNECT_IND_PLCI(&cMsg);
                connection[tmp_conid].status.sending = 1;            
                connection[tmp_conid].status.phase = CONNECT;
                connection[tmp_conid].status.bytes_received = 0;
                connection[tmp_conid].status.bytes_sent = 0;
                connection[tmp_conid].status.bytes_total = 0;
                if(ev_status) ev_status(tmp_conid, &connection[tmp_conid].status);            
            }
            capi_error(cMsg.Info);
            break;

        case CAPI_CONNECT_B3:
            dbg(2, "CNF: CAPI_CONNECT_B3");
            capi_error(cMsg.Info);
            break;
    }
}

static void capi_rx_tx(int con)
{        
    _cmsg CMSG;
    uint8_t alaw_buffer_rx[CAPI_PACKETS];
    uint8_t alaw_buffer_tx[CAPI_PACKETS];    
    int len = DATA_B3_IND_DATALENGTH(&cMsg);    
    int ret;
    
    memset(alaw_buffer_rx, 0, sizeof(alaw_buffer_rx));
    memset(alaw_buffer_tx, 0, sizeof(alaw_buffer_tx));
    
    // buffer-length check
    if(len > CAPI_PACKETS)
    {
        dbg(1, "Warning: Buffer overflow! %d > %d", len, CAPI_PACKETS);
        len = CAPI_PACKETS;
    }
    
    // <-- CAPI
    memcpy(alaw_buffer_rx, (uint8_t*)DATA_B3_IND_DATA(&cMsg), len);    
    DATA_B3_RESP(&CMSG, ApplId, 0, connection[con].NCCI, DATA_B3_IND_DATAHANDLE(&cMsg));        
    
    // RX/TX spandsp    
    ret = spandsp_rx(con, alaw_buffer_rx, len);        
    if(!ret) ret = spandsp_tx(con, alaw_buffer_tx, len);    
    
    #if SPANDSP_VERSION <= 5
    // Progress (old spandsp)
    if(connection[con].status.sending) 
    {
        if(connection[con].fax->t30_state.phase == 8 && connection[con].fax->hdlctx.byte > 0)
        {                        
            connection[con].status.bytes_sent += (connection[con].fax->hdlctx.byte / 8);
            #if SPANDSP_VERSION == 5
            connection[con].status.bytes_received = connection[con].fax->t30_state.t4.line_image_size / 8; 
            #endif

            if(connection[con].hdl % 100 == 0 && ev_progress)     // progress every 2 sec       
                ev_progress(con, &connection[con].status);
        }
    }
    else
    {
        if(connection[con].fax->t30_state.phase == 7 && connection[con].fax->hdlcrx.rx_bytes > 0) 
        {            
            connection[con].status.bytes_received = connection[con].fax->hdlcrx.rx_bytes;            

            if(connection[con].hdl % 100 == 0 && ev_progress)     // progress every 2 sec
                ev_progress(con, &connection[con].status);
        }            
    }
    #endif

    // --> CAPI
    DATA_B3_REQ(&CMSG, ApplId, 0, connection[con].NCCI, (void*)alaw_buffer_tx, len, connection[con].hdl, 0);            
    
    connection[con].hdl++;
}

static int capi_call(const char *src_no, const char *trg_no, unsigned controller, unsigned call_anonymous)
{
    _cmsg cmsg;
    unsigned char called_party_number[70];
    unsigned char calling_party_number[70];
    int err = 0;

    dbg(1, "REQ: CONNECT (%s -> %s)", src_no, trg_no);
        
    // TargetNo
    called_party_number[0] = 1 + strlen(trg_no);
    called_party_number[1] = 0x80;
    strncpy((char *) &called_party_number[2], trg_no, strlen(trg_no));

    // MSN
    calling_party_number[0] = 2 + strlen(src_no);
    calling_party_number[1] = 0x00;
    calling_party_number[2] = 0x80;
    if(call_anonymous) calling_party_number[2] = 0xA0;
    strncpy((char *) &calling_party_number[3], src_no, strlen(src_no));
        
    // Request connect
    err = CONNECT_REQ(&cmsg, ApplId,
            0,
            controller,                                 // Controller
            CAPI_CIP,                                   // CIP-value 
            (unsigned char *)called_party_number,       // TargetNo
            (unsigned char *)calling_party_number,      // Number
            (unsigned char *)"",                        // CalledPartySubaddress
            (unsigned char *)"",                        // CallingPartySubaddress,
            1,                                          // B1 transp
            1,                                          // B2 transp
            0,                                          // B3 transp
            (unsigned char *)"",                        // B1configuration
            (unsigned char *)"",                        // B2configuration
            (unsigned char *)"",                        // B3configuration
            (unsigned char *)"",                        // Globalconfig
            (unsigned char *)"",                        // BC
            (unsigned char *)"",                        // HLC
            (unsigned char *)"",                        // LLC
            (unsigned char *)"",                        // B-channel information
            (unsigned char *)"",                        // Keypadfacility
            (unsigned char *)"",                        // Useruserdata
            (unsigned char *)"",                        // Facilitydataarray
            (_cstruct)"");                              // SendingComplete

    // Error?
    if(err)
    {
        dbg(1, "(%d) Unable to send CONNECT_REQ!", err);
        capi_error(err);        
        return -1;
    }    
        
    return 0;
}

static void capi_resp_connection(int PLCI, unsigned int ignore)
{
    _cmsg CMSG1;

    if(!ignore)
    {
        // *Ring*
        ALERT_REQ(&CMSG1, ApplId, 0, PLCI,
                NULL,       
                NULL,       
                NULL,       
                NULL, 
                NULL);
    }

                            
    // Ignore
    CONNECT_RESP(&CMSG1,
             ApplId,
             0,           // MsgNr.
             PLCI,
             ignore,      // Reject
             1,           // B1 Protocol
             1,           // B2 Protocol
             0,           // B3 Protocol
             NULL,        // B1configuration
             NULL,        // B2configuration
             NULL,        // B3configuration
             NULL,        // Globalconfiguration
             NULL,        // ConnectedNumber
             NULL,        // ConnectedSubaddress
             NULL,        // LLC
             NULL,        // BChannelinformation
             NULL,        // Keypadfacility
             NULL,        // Useruserdata
             NULL);       // Facilitydataarray        
}

static void capi_term_connection(int con)
{
    _cmsg CMSG1;

    connection[con].NCCI = 0;   // set NCCI to 0, so that DATA_B3_REQ stops

    if(connection[con].PLCI > 0)
        DISCONNECT_REQ(&CMSG1, ApplId, 1, connection[con].PLCI, NULL, NULL, NULL, NULL);
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Get the calling party number on CAPI_CONNECT
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_get_source_no(_cmsg *CMSG, char number[256])
{    
    unsigned char *x=CONNECT_IND_CALLINGPARTYNUMBER(CMSG);
    unsigned int len = 0;

    memset(number, 0, sizeof(number));

    if(x == NULL) {
        x = INFO_IND_INFOELEMENT(CMSG);
        if(x != NULL)
            len = (int)x[0];
    }
    else {
        len = *CONNECT_IND_CALLINGPARTYNUMBER(CMSG);
    }
                
    if(len <= 1)
    {
        strcpy(number, "unknown");
    }
    else
    {
        if(len>sizeof(number))
            len = sizeof(number)-1;

        // get number
        if(x[2] & 128)
        {
           number[strlen(number)+x[0]-1]=0;
           number[strlen(number)+x[0]-2]=0;
           memcpy(number+strlen(number),x+3,(size_t)(x[0]-2));
        }
        else
        {
           number[strlen(number)+x[0]]=0;
           number[strlen(number)+x[0]-1]=0;
           memcpy(number+strlen(number),x+2,(size_t)(x[0]-1));
        }
    }

    if(!strlen(number))
        strcpy(number, "anonymous");    
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Get the called party number on CAPI_CONNECT
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_get_target_no(_cmsg *CMSG, char number[256])
{
    unsigned char *x=CONNECT_IND_CALLEDPARTYNUMBER(CMSG);
    unsigned int len = 0;

    memset(number, 0, sizeof(number));
    
    if(x == NULL) {
        x = INFO_IND_INFOELEMENT(CMSG);
        if(x != NULL)
            len = (int)x[0];
    }
    else {
        len = *CONNECT_IND_CALLEDPARTYNUMBER(CMSG);
        if(CONNECT_IND_CALLEDPARTYNUMBER(CMSG)[0] == 0)
            len = 0;
    }

    if(len <= 1)
    {
        strcpy(number, "unknown");
    }
    else
    {
        if(len>sizeof(number))
            len = sizeof(number)-1;

        // get number
        number[strlen(number)+(size_t)x[0]]=0;
        number[strlen(number)+(size_t)x[0]-1]=0;
        memcpy(number+strlen(number),x+2,(size_t)(x[0]-1));
    }

    if(!strlen(number))
        strcpy(number, "anonymous");    
}



static void capi_error(long info)
{
    char capi_err[64];

    snprintf(capi_err, sizeof(capi_err), "CAPI 0x%lx -", info);

    switch(info)
    {
        // default capi
        case 0x0001:
            dbg(1, "%s NCPI not supported by current protocol", capi_err);
            break;
        case 0x0002:
            dbg(1, "%s Flags not supported by current protocol, flags ignored", capi_err);
            break;
        case 0x0003:
            dbg(1, "%s Alert already sent by another application", capi_err);
            break;

        // custom
        case 0x0999:
            dbg(1, "%s Rejected by user!", capi_err);
            break;
        case 0x1000:
            dbg(1, "%s Aborted by user!", capi_err);
            break;
        
        // default capi
        case 0x1001:
            dbg(1, "%s Too many applications", capi_err);
            break;
        case 0x1002:
            dbg(1, "%s Logical Block size too small; must be at least 128 bytes", capi_err);
            break;
        case 0x1003:
            dbg(1, "%s Buffer exceeds 64 kbytes", capi_err);
            break;
        case 0x1004:
            dbg(1, "%s Message buffer size too small, must be at least 1024 bytes", capi_err);
            break;
        case 0x1005:
            dbg(1, "%s Max. number of logical connections not supported", capi_err);
            break;
        case 0x1006:
            dbg(1, "%s reserved (unknown error)", capi_err);
            break;
        case 0x1007:
            dbg(1, "%s The message could not be accepted because of an internal busy condition", capi_err);
            break;
        case 0x1008:
            dbg(1, "%s OS Resource error (out of memory?)", capi_err);
            break;
        case 0x1009:
            dbg(1, "%s CAPI not installed", capi_err);
            break;
        case 0x100A:
            dbg(1, "%s Controller does not support external equipment", capi_err);
            break;
        case 0x100B:
            dbg(1, "%s Controller does only support external equipment", capi_err);
            break;
        case 0x1101:
            dbg(1, "%s Illegal application number", capi_err);
            break;
        case 0x1102:
            dbg(1, "%s Illegal command or subcommand, or message length less than 12 octets", capi_err);
            break;
        case 0x1103:
            dbg(1, "%s The message could not be accepted because of a queue full condition", capi_err);
            break;
        case 0x1104:
            dbg(1, "%s Queue overflow: a message was lost!!", capi_err);
            break;
        case 0x1105:
            dbg(1, "%s Unknown notification parameter", capi_err);
            break;
        case 0x1106:
            dbg(1, "%s The message could not be accepted because on an internal busy condition", capi_err);
            break;
        case 0x1107:
            dbg(1, "%s OS resource error (out of memory?)", capi_err);
            break;
        case 0x1108:
            dbg(1, "%s CAPI not installed", capi_err);
            break;
        case 0x1109:
            dbg(1, "%s Controller does not support external equipment", capi_err);
            break;
        case 0x110A:
            dbg(1, "%s Controller does only support external equipment", capi_err);
            break;
        case 0x110B:
            dbg(1, "%s Controller does only support external equipment", capi_err);
            break;
        case 0x2001:
            dbg(1, "%s Message not supported in current state", capi_err);
            break;
        case 0x2002:
            dbg(1, "%s Illegal Controller / PLCI / NCCI", capi_err);
            break;
        case 0x2003:
            dbg(1, "%s Out of PLCI", capi_err);
            break;
        case 0x2004:
            dbg(1, "%s Out of NCCI", capi_err);
            break;
        case 0x2005:
            dbg(1, "%s Out of LISTEN", capi_err);
            break;
        case 0x2007:
            dbg(1, "%s llegal message parameter coding", capi_err);
            break;
        case 0x3001:
            dbg(1, "%s B1 protocol not supported", capi_err);
            break;
        case 0x3002:
            dbg(1, "%s B2 protocol not supported", capi_err);
            break;
        case 0x3003:
            dbg(1, "%s B3 protocol not supported", capi_err);
            break;
        case 0x3004:
            dbg(1, "%s B1 protocol parameter not supported", capi_err);
            break;
        case 0x3005:
            dbg(1, "%s B2 protocol parameter not supported", capi_err);
            break;
        case 0x3006:
            dbg(1, "%s B3 protocol parameter not supported", capi_err);
            break;
        case 0x3007:
            dbg(1, "%s B protocol combination not supported", capi_err);
            break;
        case 0x3008:
            dbg(1, "%s NCPI not supported", capi_err);
            break;
        case 0x3009:
            dbg(1, "%s CIP Value unknown", capi_err);
            break;
        case 0x300A:
            dbg(1, "%s Flags not supported (reserved bits)", capi_err);
            break;
        case 0x300B:
            dbg(1, "%s Facility not supported", capi_err);
            break;
        case 0x300C:
            dbg(1, "%s Data length not supported by current protocol", capi_err);
            break;
        case 0x300D:
            dbg(1, "%s Reset procedure not supported by current protocol", capi_err);
            break;

        // informations about the clearing of a physical connection
        case 0x3301:
            dbg(1, "%s Protocol error layer 1 (broken line or B-channel removed by signalling protocol)", capi_err);
            break;
        case 0x3302:
            dbg(1, "%s Protocol error layer 2", capi_err);
            break;
        case 0x3303:
            dbg(1, "%s Protocol error layer 3", capi_err);
            break;
        case 0x3304:
            dbg(1, "%s Another application got that call", capi_err);
            break;

        // disconnect errors / reasons
        case 0x3400:
            dbg(1, "%s Normal termination, no reason available", capi_err);        
            break;
        case 0x3480:
            dbg(1, "%s Normal termination", capi_err);        
            break;
        case 0x3481:
            dbg(1, "%s Unallocated (unassigned) number", capi_err);        
            break;
        case 0x3482:
            dbg(1, "%s No route to specified transit network", capi_err);
            break;
        case 0x3483:
            dbg(1, "%s No route to destination", capi_err);
            break;
        case 0x3486:
            dbg(1, "%s Channel unacceptable", capi_err);
            break;
        case 0x3487:
            dbg(1, "%s Call awarded and being delivered in an established channel", capi_err);
            break;
        case 0x3490:
            dbg(1, "%s Normal call clearing", capi_err);
            break;
        case 0x3491:
            dbg(1, "%s User busy", capi_err);
            break;
        case 0x3492:
            dbg(1, "%s No user responding", capi_err);
            break;
        case 0x3493:
            dbg(1, "%s No answer from user (user alerted)", capi_err);
            break;
        case 0x3495:
            dbg(1, "%s Call rejected", capi_err);
            break;
        case 0x3496:
            dbg(1, "%s Number changed", capi_err);
            break;
        case 0x349A:
            dbg(1, "%s Non-selected user clearing", capi_err);
            break;
        case 0x349B:
            dbg(1, "%s Destination out of order", capi_err);
            break;
        case 0x349C:
            dbg(1, "%s Invalid number format", capi_err);
            break;
        case 0x349D:
            dbg(1, "%s Facility rejected", capi_err);
            break;
        case 0x349E:
            dbg(1, "%s Response to STATUS ENQUIRY", capi_err);
            break;
        case 0x349F:
            dbg(1, "%s Normal, unspecified", capi_err);
            break;
        case 0x34A2:
            dbg(1, "%s No circuit / channel available", capi_err);
            break;
        case 0x34A6:
            dbg(1, "%s Network out of order", capi_err);
            break;
        case 0x34A9:
            dbg(1, "%s Temporary failure", capi_err);
            break;
        case 0x34AA:
            dbg(1, "%s Switching equipment congestion", capi_err);
            break;
        case 0x34AB:
            dbg(1, "%s Access information discarded", capi_err);
            break;
        case 0x34AC:
            dbg(1, "%s Requested circuit / channel not available", capi_err);
            break;
        case 0x34AF:
            dbg(1, "%s Resources unavailable, unspecified", capi_err);
            break;
        case 0x34B1:
            dbg(1, "%s Quality of service unavailable", capi_err);
            break;
        case 0x34B2:
            dbg(1, "%s Requested facility not subscribed", capi_err);
            break;
        case 0x34B9:
            dbg(1, "%s Bearer capability not authorized", capi_err);
            break;
        case 0x34BA:
            dbg(1, "%s Bearer capability not presently available", capi_err);
            break;
        case 0x34BF:
            dbg(1, "%s Service or option not available, unspecified", capi_err);
            break;
        case 0x34C1:
            dbg(1, "%s Bearer capability not implemented", capi_err);
            break;
        case 0x34C2:
            dbg(1, "%s Channel type not implemented", capi_err);
            break;
        case 0x34C5:
            dbg(1, "%s Requested facility not implemented", capi_err);
            break;
        case 0x34C6:
            dbg(1, "%s Only restricted digital information bearer capability is available", capi_err);
            break;
        case 0x34CF:
            dbg(1, "%s Service or option not implemented, unspecified", capi_err);
            break;
        case 0x34D1:
            dbg(1, "%s Invalid call reference value", capi_err);
            break;
        case 0x34D2:
            dbg(1, "%s Identified channel does not exist", capi_err);
            break;
        case 0x34D3:
            dbg(1, "%s A suspended call exists, but this call identity does not", capi_err);
            break;
        case 0x34D4:
            dbg(1, "%s Call identity in use", capi_err);
            break;
        case 0x34D5:
            dbg(1, "%s No call suspended", capi_err);
            break;
        case 0x34D6:
            dbg(1, "%s Call having the requested call identity has been cleared", capi_err);
            break;
        case 0x34D8:
            dbg(1, "%s Incompatible destination", capi_err);
            break;
        case 0x34DB:
            dbg(1, "%s Invalid transit network selection", capi_err);
            break;
        case 0x34DF:
            dbg(1, "%s Invalid message, unspecified", capi_err);
            break;
        case 0x34E0:
            dbg(1, "%s Mandatory information element is missing", capi_err);
            break;
        case 0x34E1:
            dbg(1, "%s Message type non-existent or not implemented", capi_err);
            break;
        case 0x34E2:
            dbg(1, "%s Message not compatible with call state or message type non-existent or not implemented", capi_err);
            break;
        case 0x34E3:
            dbg(1, "%s Information element non-existent or not implemented", capi_err);
            break;
        case 0x34E4:
            dbg(1, "%s Invalid information element contents", capi_err);
            break;
        case 0x34E5:
            dbg(1, "%s Message not compatible with call state", capi_err);
            break;
        case 0x34E6:
            dbg(1, "%s Recovery on timer expiry", capi_err);
            break;
        case 0x34EF:
            dbg(1, "%s Protocol error, unspecified", capi_err);
            break;
        case 0x34FF:
            dbg(1, "%s Interworking, unspecified", capi_err);
            break;
    }
}

static int capi_get_free()
{
    int ret = -1;
    int i;
    
    for(i=0; i<CAPI_CONNECTIONS; i++)
    {
        if(connection[i].PLCI == 0 && connection[i].NCCI == 0) {
            ret = i;
            break;
        }
    }
    
    return ret;
}

static int capi_set_free(int con)
{
    int i;
    int in_use = 0;
    
    // reset connection        
    connection[con].NCCI = 0;
    connection[con].hdl = 0;

    if(!connection[con].status.sending && in_thread) connection[con].PLCI = -1;
    else connection[con].PLCI = 0;

    // check, if we should stay in listen loop (parallel receive/send)
    for(i=0; i<CAPI_CONNECTIONS; i++)
    {
        if((connection[i].PLCI != 0 || connection[i].NCCI != 0) && connection[con].PLCI != -1)
        {
            in_use = 1;
            break;
        }
    }
                    
    // Exit loop and reschedule process priority
    if(!in_use)         
        boost_priority(0);
        
    // Force PHASE_E (!)    
    if(connection[con].status.phase <= PHASE_D)
    {
        for(i=0; i<32768; i++)
        {
            uint8_t buf[CAPI_PACKETS];            
            memset(buf, 128, CAPI_PACKETS);
            spandsp_rx(con, buf, CAPI_PACKETS);
            spandsp_tx(con, buf, CAPI_PACKETS);

            if(connection[con].status.phase == PHASE_E)
                break;
        }
    }

    // Free fax
    spandsp_close(con);

    // Fax not ok???
    if(connection[con].status.phase != FAX_OK && connection[con].status.phase != FAX_REJECTED)
        connection[con].status.phase = FAX_FAILED;      // Transmit/Receive-Error!

    // Last (custom) phase (ok=3/error=4)
    if(ev_status)    
        ev_status(con, &connection[con].status);
        
    // Reinit Fax (stay in loop / thread)    
    if(!connection[con].status.sending)
        spandsp_init(connection[con].status.tiff_file, connection[con].status.sending, connection[con].status.modem, connection[con].status.ecm, connection[con].status.ident, connection[con].status.header, con);

    // reset some variables    
    connection[con].status.phase = IDLE;
    connection[con].status.bytes_received = 0;
    connection[con].status.bytes_sent = 0;
    connection[con].status.bytes_total = 0;
    connection[con].status.bad_rows = 0;
    connection[con].status.bitrate = 0;
    connection[con].status.capi_code = 0;

    // Not in thread? Then exit capi loop and return error code!
    if(!in_thread) 
        in_loop = 0;
    else 
        connection[con].status.error_code = 0;        
    
    return 0;
}

static int capi_find_plci(int PLCI)
{
    int ret = -1;
    int i;
    
    for(i=0; i<CAPI_CONNECTIONS; i++)
    {
        if(connection[i].PLCI == PLCI) {
            ret = i;
            break;
        }
    }
    
    return ret;
}

static int capi_find_ncci(int NCCI)
{
    int ret = -1;
    int i;
    
    for(i=0; i<CAPI_CONNECTIONS; i++)
    {
        if(connection[i].NCCI == NCCI) {
            ret = i;
            break;
        }
    }
    
    return ret;
}

/* --------------------------------------------------------
    spandsp functions
-------------------------------------------------------- */
#if SPANDSP_VERSION >= 6
static void real_time_frame_handler(t30_state_t *s, void *user_data, int direction, const uint8_t *msg, int len)
{
    t30_stats_t t;    
    int i = (intptr_t) user_data;
    
    if(msg[2] == 6)
    {   
        t30_get_transfer_statistics (s, &t);

        if(connection[i].status.sending) {                
            connection[i].status.bytes_total = t.image_size;
            connection[i].status.bytes_sent += len;             
        } else {
            connection[i].status.bytes_received += len;            
            connection[i].status.bytes_total += len;
        }        
                
        ev_progress(i, &connection[i].status);        
    }
}
#endif

static int spandsp_init(const char *tiff_file, int sending, int modem, int ecm, const char *LSI, const char *LocalHeaderInfo, int conid)
{   
    t30_state_t *t30 = NULL;
    logging_state_t *logging;
    long con = -1;
    int supported_compressions = 0;
    int supported_resolutions = 0;
    int supported_image_sizes = 0;
    int supported_modems = 0;

    if(conid == -1) con = capi_get_free();
    else con = conid;

    if(con == -1) return con;    
    
    connection[con].fax = fax_init(connection[con].fax, sending);        

    fax_set_transmit_on_idle(connection[con].fax, TRUE);    
    fax_set_tep_mode(connection[con].fax, TRUE);

    #if SPANDSP_VERSION >= 6
    t30 = fax_get_t30_state(connection[con].fax);
    #else
    t30 = &connection[con].fax->t30_state;
    #endif
        
    // Supported compressions
    supported_compressions = 0;
    supported_compressions |= T30_SUPPORT_NO_COMPRESSION;        
    supported_compressions |= T30_SUPPORT_T4_1D_COMPRESSION;
    supported_compressions |= T30_SUPPORT_T4_2D_COMPRESSION; 
    if(ecm) supported_compressions |= T30_SUPPORT_T6_COMPRESSION;
    
    // Supported resolutions
    supported_resolutions = 0;
    supported_resolutions |= T30_SUPPORT_STANDARD_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_FINE_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_SUPERFINE_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_R4_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_R8_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_R16_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_300_300_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_400_400_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_600_600_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_1200_1200_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_300_600_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_400_800_RESOLUTION;
    supported_resolutions |= T30_SUPPORT_600_1200_RESOLUTION;

    // Supported image sizes
    supported_image_sizes = 0;
    supported_image_sizes |= T30_SUPPORT_215MM_WIDTH;
    supported_image_sizes |= T30_SUPPORT_255MM_WIDTH;
    supported_image_sizes |= T30_SUPPORT_303MM_WIDTH;
    supported_image_sizes |= T30_SUPPORT_UNLIMITED_LENGTH;
    supported_image_sizes |= T30_SUPPORT_A4_LENGTH;
    supported_image_sizes |= T30_SUPPORT_B4_LENGTH;
    supported_image_sizes |= T30_SUPPORT_US_LETTER_LENGTH;
    supported_image_sizes |= T30_SUPPORT_US_LEGAL_LENGTH;
    
    // Supported modems
    supported_modems = 0;
    if (modem > 0) supported_modems |= T30_SUPPORT_V27TER;
    if (modem > 1) supported_modems |= T30_SUPPORT_V29;
    if (modem > 2) supported_modems |= T30_SUPPORT_V17;        
    t30_set_supported_modems(t30, supported_modems);

    // Error correction
    if(ecm) t30_set_ecm_capability (t30, TRUE);    
        
    t30_set_supported_compressions(t30, supported_compressions);
    t30_set_supported_resolutions(t30, supported_resolutions);
    t30_set_supported_image_sizes(t30, supported_image_sizes);
        
    // spandsp loglevel
    if(log_level >= 3)
    {
        #if SPANDSP_VERSION >= 6
        logging = t30_get_logging_state(t30);
        span_log_set_level(logging, 0xFFFFFF);                        
        #else
        logging = &t30->logging;
        logging->level = 0xFFFFFF;        
        #endif

        span_log_set_message_handler(logging, spandsp_msg_log);
    }
    
    #if SPANDSP_VERSION >= 5
    if (LSI) t30_set_tx_ident (t30, LSI);
    if (LocalHeaderInfo) t30_set_tx_page_header_info (t30, LocalHeaderInfo);
    #endif
    #if SPANDSP_VERSION == 4
    if (LSI) t30_set_local_ident (t30, LSI);
    if (LocalHeaderInfo) t30_set_header_info (t30, LocalHeaderInfo);
    #endif
   
    if (sending == TRUE)
        t30_set_tx_file (t30, tiff_file, -1, -1);
    else 
        t30_set_rx_file (t30, tiff_file, -1);
        
    t30_set_phase_b_handler (t30, phase_b_handler, (void*)con);
    t30_set_phase_d_handler (t30, phase_d_handler, (void*)con);
    t30_set_phase_e_handler (t30, phase_e_handler, (void*)con);    

    #if SPANDSP_VERSION >= 6
    t30_set_real_time_frame_handler(t30, real_time_frame_handler, (void *)con);
    #endif

    return con;
} 

static int spandsp_close(int con)
{
    int i;    
    
    if(con < 0) {
        for(i=0; i<CAPI_CONNECTIONS; i++) {
            if(connection[i].fax != NULL)
                fax_release(connection[i].fax);          
        }
    } else {        
        if(connection[con].fax != NULL)
            fax_release(connection[con].fax);
    }
        
    return 0;
}

static int spandsp_tx(int con, uint8_t *buf, size_t len)
{    
    int16_t buf_in[CAPI_PACKETS];    
    uint8_t *alaw;
    int err, i;
        
    err = fax_tx(connection[con].fax, buf_in, CAPI_PACKETS);
    alaw = buf;
    
    // convert (pcm -> alaw/ulaw)           
    for (i=0; i!=len; ++i, ++alaw) 
        *alaw = _linear16_2_law[(int16_t)buf_in[i]];          
            
    return err;
}

static int spandsp_rx(int con, uint8_t *buf, size_t len)
{
    int16_t buf_in[CAPI_PACKETS];
    int16_t *wave;
    int err, i;
        
    wave = buf_in;

    // convert (alaw/ulaw -> pcm)
    for (i=0; i!=len; ++i, ++wave) 
        *wave = _law_2_linear16[(uint8_t)buf[i]];          

    err = fax_rx(connection[con].fax, buf_in, CAPI_PACKETS);
    
    return err;
}

#if SPANDSP_VERSION >= 5
static int phase_b_handler(t30_state_t *s, void *user_data, int result)
#else
static void phase_b_handler(t30_state_t *s, void *user_data, int result)
#endif
{
    t30_stats_t t;
    t30_state_t *t30;
    const char *pfx = "Phase B:";    
    long con_id = (long)(user_data);
    
    t30_get_transfer_statistics (s, &t);
    #if SPANDSP_VERSION >= 6
    t30 = fax_get_t30_state(connection[con_id].fax);
    #else
    t30 = &connection[con_id].fax->t30_state;
    #endif

    dbg(1, "Phase B handler (0x%X) %s", result, t30_frametype(result));
    dbg(1, "%s bit rate %d", pfx, t.bit_rate);
    dbg(1, "%s ECM %s", pfx, (t.error_correcting_mode)  ?  "on"  :  "off");    
    
    #if SPANDSP_VERSION >= 6
    if(connection[con_id].status.sending) snprintf(connection[con_id].status.ident, sizeof(connection[con_id].status.ident), "%s", t30_get_tx_ident(t30));    
    else snprintf(connection[con_id].status.ident, sizeof(connection[con_id].status.ident), "%s", t30_get_rx_ident(t30));
    #endif

    #if SPANDSP_VERSION == 5
    if(!connection[con_id].status.sending) snprintf(connection[con_id].status.ident, sizeof(connection[con_id].status.ident), "%s", connection[con_id].fax->t30_state.rx_info.ident);
    else snprintf(connection[con_id].status.ident, sizeof(connection[con_id].status.ident), "%s", connection[con_id].fax->t30_state.tx_info.ident);
    #endif

    connection[con_id].status.phase = PHASE_B;
    connection[con_id].status.bytes_sent = 0;
    connection[con_id].status.bytes_total = 0;
    connection[con_id].status.bytes_received = 0;    
    connection[con_id].status.ecm = t.error_correcting_mode;
    connection[con_id].status.bad_rows = t.bad_rows;
    connection[con_id].status.encoding = t.encoding;
    connection[con_id].status.bitrate = t.bit_rate;
    connection[con_id].status.page_total =  t.pages_in_file;    

    #if SPANDSP_VERSION >= 6
    connection[con_id].status.page_current = connection[con_id].status.sending ? t.pages_tx + 1 : t.pages_rx + 1;
    #else
    connection[con_id].status.page_current = t.pages_transferred + 1;
    #endif
        
    if(ev_status) ev_status(con_id, &connection[con_id].status);

    #if SPANDSP_VERSION >= 5
    return 0;
    #endif
}

#if SPANDSP_VERSION >= 5
static int phase_d_handler(t30_state_t *s, void *user_data, int result)
#else
static void phase_d_handler(t30_state_t *s, void *user_data, int result)
#endif
{    
    const char *pfx = "Phase D:";
    t30_stats_t t;
    long con_id = (long)(user_data);

    t30_get_transfer_statistics(s, &t);

    dbg(1, "Phase D handler (0x%X) %s", result, t30_frametype(result));
    #if SPANDSP_VERSION >= 6
    dbg(1, "%s pages transferred %d", pfx, connection[con_id].status.sending ? t.pages_tx : t.pages_rx);
    #else
    dbg(1, "%s pages transferred %d", pfx, t.pages_transferred);
    #endif
    dbg(1, "%s image size %d x %d", pfx, t.width, t.length);    
    dbg(1, "%s bad rows %d", pfx, t.bad_rows);
    dbg(1, "%s longest bad row run %d", pfx, t.longest_bad_row_run);
    dbg(1, "%s image size %d", pfx, t.image_size);
    
    connection[con_id].status.phase = PHASE_D;
    connection[con_id].status.ecm = t.error_correcting_mode;
    connection[con_id].status.bad_rows = t.bad_rows;
    connection[con_id].status.encoding = t.encoding;
    connection[con_id].status.bitrate = t.bit_rate;

    #if SPANDSP_VERSION >= 6
    if(connection[con_id].status.sending)
        connection[con_id].status.page_current = (t.pages_in_file >= t.pages_tx + 1 ? t.pages_tx + 1 : t.pages_tx); 
    else
        connection[con_id].status.page_current = t.pages_rx; 
    #else
        connection[con_id].status.page_current = (t.pages_in_file >= t.pages_transferred + 1 ? t.pages_transferred + 1 : t.pages_transferred); 
    #endif

    connection[con_id].status.page_total = t.pages_in_file;
    connection[con_id].status.bytes_received = 0;

    if(ev_status) ev_status(con_id, &connection[con_id].status);
    
    #if SPANDSP_VERSION >= 5
    return 0;
    #endif
}

static void phase_e_handler(t30_state_t *s, void *user_data, int result)
{    
    long con_id = -1;    
    const char *pfx = "Phase E:";
    t30_stats_t t;
    t30_state_t *t30;
        
    con_id = (long)(user_data);
    
    t30_get_transfer_statistics(s, &t);

    dbg(1, "Phase E handler (0x%X) %s", result, t30_completion_code_to_str(result));
    #if SPANDSP_VERSION >= 6
    dbg(1, "%s pages transferred %d", pfx, connection[con_id].status.sending ? t.pages_tx : t.pages_rx);
    #else
    dbg(1, "%s pages transferred %d", pfx, t.pages_transferred);
    #endif
    dbg(1, "%s image resolution %d x %d", pfx, t.x_resolution, t.y_resolution);
    dbg(1, "%s compression type %d", pfx, t.encoding);
    dbg(1, "%s coding method %s", pfx, t4_encoding_to_str(t.encoding));

    connection[con_id].status.phase = PHASE_E;    
    connection[con_id].status.ecm = t.error_correcting_mode;
    connection[con_id].status.bad_rows = t.bad_rows;
    connection[con_id].status.encoding = t.encoding;
    connection[con_id].status.bitrate = t.bit_rate;
    #if SPANDSP_VERSION >= 6
    connection[con_id].status.page_current = (connection[con_id].status.sending ? t.pages_tx : t.pages_rx);    
    #else
    connection[con_id].status.page_current = t.pages_transferred;
    #endif
    connection[con_id].status.page_total = t.pages_in_file;
    connection[con_id].status.error_code = result;
    
    #if SPANDSP_VERSION >= 6
    t30 = fax_get_t30_state(connection[con_id].fax);
    if(connection[con_id].status.sending) snprintf(connection[con_id].status.ident, sizeof(connection[con_id].status.ident), "%s", t30_get_tx_ident(t30));    
    else snprintf(connection[con_id].status.ident, sizeof(connection[con_id].status.ident), "%s", t30_get_rx_ident(t30));
    #endif

    #if SPANDSP_VERSION == 5
    t30 = &connection[con_id].fax->t30_state;
    if(!connection[con_id].status.sending) snprintf(connection[con_id].status.ident, sizeof(connection[con_id].status.ident), "%s", connection[con_id].fax->t30_state.rx_info.ident);
    else snprintf(connection[con_id].status.ident, sizeof(connection[con_id].status.ident), "%s", connection[con_id].fax->t30_state.tx_info.ident);
    #endif

    if(ev_status) ev_status(con_id, &connection[con_id].status);    
    if(result == 0) connection[con_id].status.phase = FAX_OK;
    
    // disconnect (only when sending)
    if(connection[con_id].status.sending > 0)
        capi_term_connection(con_id);
}

static void spandsp_msg_log(int level, const char *text)
{
    dbg(3, text);
}


/* --------------------------------------------------------
    other functions
-------------------------------------------------------- */
static void dbg(int level, const char *format, ...)
{        
    if(level <= log_level)
    {        
        char output[512];
        char output2[700];   

        va_list args;        
        time_t tim = time(NULL);
        struct tm *now_t = localtime(&tim);        

        va_start(args, format);
        vsnprintf(output, sizeof(output), format, args);
        snprintf(output2, sizeof(output2), "[%s] %02d.%02d.%d %02d:%02d:%02d - %s", LIBNAME, now_t->tm_mday, now_t->tm_mon+1, now_t->tm_year+1900, now_t->tm_hour, now_t->tm_min, now_t->tm_sec, output);

        if(ev_log) ev_log(output2);
        else printf("%s\n", output2);
       
        va_end(args);    
    }    
}

#if (WITH_BOOST == 1)
#ifndef WIN32
static void boost_priority(int boost)
{
    // already boosted? abort!
    if(boosted && boost) return;
    
    if(boost) 
        setpriority(PRIO_PROCESS, 0, -10);
    else
        setpriority(PRIO_PROCESS, 0, +10);

    boosted = boost;
}
#else
static void boost_priority(int boost)
{
    if(boost)
    {
        // higher priority
        SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
        SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);       
    }
    else {
        // lower priority
        SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);
        SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);        
    } 
}
#endif
#else
static void boost_priority(int boost) { }
#endif
